;;;; R to common-lisp translator: example R code
;;;; benchmark function in R, see http://dan.corlan.net/bench.html
;;;;
;;;; Copyright (c) 2005--2006 Alexandru Dan Corlan
;;;; this code is released under the GNU General Public Licence V2.
;;;; (see http://www.gnu.org/copyleft/gpl.html for details)
;;;;
;;;; Homepage: http://dan.corlan.net/R_to_common_lisp_translator
;;;;
;;;; Usage: R2cl filename
;;;;
;;;; will translate filename.R into filename.lisp which
;;;; requires rruntime.x86f (compiled from rruntime.lisp)
;;;;
;;;; V0.1: Jan 7, 2006, minimal proof of concept

(provide "rruntime")
(setf *gc-verbose* nil)

(defun seq (from to &optional (by 1))
  (declare (fixnum from) (fixnum to) (fixnum by))
  (let* ((l (the fixnum (+ (/ (the fixnum (- to from)) by) 1)))
	 (r (make-array l :element-type 'single-float))
	 )
    (dotimes (i l)
      (setf (elt r i)
	    (coerce (the fixnum (+ from (the fixnum (* (the fixnum i) by))))
		    'single-float)))
    r))

(defun dcl-to-let (l)
  (if (and (consp l) (consp (cdr l)))
      (cons (list (cadr l) (if (eq (car l) 'scalar) 0.0 nil))
	    (dcl-to-let (cddr l)))
    nil))
(defun dcl-to-declare (l)
  (if (and (consp l) (consp (cdr l)))
      (cond
       ((eq (car l) 'scalar)
	(cons (list 'single-float (cadr l)) (dcl-to-declare (cddr l))))
       ((eq (car l) 'int)
	(cons (list 'fixnum (cadr l)) (dcl-to-declare (cddr l))))
       (t (dcl-to-declare (cddr l)))
       )
    nil))

(defmacro R-progn (&rest dcl)
  	 (if (and (consp dcl)
		  (consp (car dcl))
		  (eq (caar dcl) 'r-let))
	     (list 'let
		   (dcl-to-let (cdar dcl))
		   (cons 'declare (dcl-to-declare (cdar dcl)))
		   (cons 'progn (cdr dcl)))
	   (cons 'progn dcl))
  )

;; defun will be something like
;; (<- name (function (llist) (R-progn [(R-declare m n ...)] [(R-let m n ...)]) body))
;;
;; which must be converted into something like
;;
;; (defun name llist [(declare m n ...)] (R-progn [(R-let ...)] body))

(defun convert-decl (ll) ;something like ((R-progn [(R-declare ...)] [(R-let ...)] body))
  (let ((l (car ll)))
    (if (and (consp (cadr l)) (eq (caadr l) 'R-declare))
	(list (cons 'declare (dcl-to-declare (cdadr l)))
	      (cons 'R-progn (cddr l)))
      l)
    )
  )

(defmacro <- (name value)
    
  (if (and (consp value) (eq (car value) 'function))
      (cons 'defun
	    (cons name
		  (cons (cadr value) ; lambda list
			(convert-decl (cddr value)))))
    (list 'setf name
	  (list 'if (list 'numberp value)
		(list 'coerce value ''single-float)
		value)
	  )
    ))

(defmacro <<- (name value)
  (if (and (consp value) (eq (car value) 'function))
      (cons 'defun
	    (cons name (cdr value)))
	(list 'setf name
	      (list 'if (list 'numberp value)
		    (list 'coerce value ''single-float)
		    value)
	      )
    ))


;;; v+ and sv+ are experimental functions, not used

(defun v+ (x y)
  (let* ((l (length x))
	 (r (make-array l :element-type 'fixnum)))
    (dotimes (i l)
	    (declare (fixnum i))
	    (setf (aref r i) (the fixnum
			       (+ (the fixnum (aref x i))
				  (the fixnum (aref y i)))))
	    )
    r))


(defun sv+ (x y)
  (declare (fixnum x))
  (let* ((l (length y))
	 (r (make-array l :element-type 'fixnum)))
    (dotimes (i l)
	    (declare (fixnum i))
	    (setf (aref r i) (the fixnum
			       (+ x
				  (the fixnum (aref y i)))))
	    )
    r))

(declaim (inline R-op R+ R- R* R/))

;;; R-op is not used anymore

(defun R-op (op v1 v2)
  (cond
   ((and (numberp v1) (numberp v2))
      (the single-float (funcall op (the single-float v1) (the single-float v2))))
   ((and (numberp v1) (vectorp v2))
    (map 'vector op (make-array (length v2) :initial-element v1) v2))
   ((and (numberp v2) (vectorp v1))
    (map 'vector op (make-array (length v1) :initial-element v2) v1))
   ((and (vectorp v2) (vectorp v1))
    (cond
     ((eql (length v1) (length v2))
      (map 'vector op v1 v2))
     (t (error "No vector length adjustment yet"))
     ))
   (t (error "Bad operation data"))
   )
  )

(defun R+ (v1 v2)
  (cond
   ((and (fixnump v1) (fixnump v2))
      (coerce 'single-float (the integer (+ (the fixnum v1) (the fixnum v2)))))
   ((and (fixnump v1) (floatp v2))
       (the single-float (+ (the fixnum v1) (the single-float v2))))
   ((and (floatp v1) (fixnump v2))
      (the single-float (+ (the single-float v1) (the fixnum v2))))
   ((and (floatp v1) (floatp v2))
      (the single-float (+ (the single-float v1) (the single-float v2))))
   ((and (numberp v1) (numberp v2))
      (the single-float (+ (the single-float v1) (the single-float v2))))
   ((and (numberp v1) (vectorp v2))
    (map 'vector #'+ (make-array (length v2) :initial-element v1) v2))
   ((and (numberp v2) (vectorp v1))
    (map 'vector #'+ (make-array (length v1) :initial-element v2) v1))
   ((and (vectorp v2) (vectorp v1))
    (cond
     ((eql (length v1) (length v2))
      (map 'vector #'+ v1 v2))
     (t (error "No vector length adjustment yet"))
     ))
   (t (error "Bad operation data"))
   )
  )

(defun R- (v1 &optional (v2 nil))
  (when (not v2)
    (return-from R- (R* -1.0 v2)))
  (cond
   ((and (fixnump v1) (fixnump v2))
      (coerce 'single-float (the integer (- (the fixnum v1) (the fixnum v2)))))
   ((and (fixnump v1) (floatp v2))
       (the single-float (- (the fixnum v1) (the single-float v2))))
   ((and (floatp v1) (fixnump v2))
      (the single-float (- (the single-float v1) (the fixnum v2))))
   ((and (floatp v1) (floatp v2))
      (the single-float (- (the single-float v1) (the single-float v2))))
   ((and (numberp v1) (numberp v2))
      (the single-float (- (the single-float v1) (the single-float v2))))
   ((and (numberp v1) (vectorp v2))
    (map 'vector #'- (make-array (length v2) :initial-element v1) v2))
   ((and (numberp v2) (vectorp v1))
    (map 'vector #'- (make-array (length v1) :initial-element v2) v1))
   ((and (vectorp v2) (vectorp v1))
    (cond
     ((eql (length v1) (length v2))
      (map 'vector #'- v1 v2))
     (t (error "No vector length adjustment yet"))
     ))
   (t (error "Bad operation data"))
   )
  )
(defun R* (v1 v2)
  (cond
   ((and (fixnump v1) (fixnump v2))
      (coerce 'single-float (the integer (* (the fixnum v1) (the fixnum v2)))))
   ((and (fixnump v1) (floatp v2))
       (the single-float (* (the fixnum v1) (the single-float v2))))
   ((and (floatp v1) (fixnump v2))
      (the single-float (* (the single-float v1) (the fixnum v2))))
   ((and (floatp v1) (floatp v2))
      (the single-float (* (the single-float v1) (the single-float v2))))
   ((and (numberp v1) (numberp v2))
      (the single-float (* (the single-float v1) (the single-float v2))))
   ((and (numberp v1) (vectorp v2))
    (map 'vector #'* (make-array (length v2) :initial-element v1) v2))
   ((and (numberp v2) (vectorp v1))
    (map 'vector #'* (make-array (length v1) :initial-element v2) v1))
   ((and (vectorp v2) (vectorp v1))
    (cond
     ((eql (length v1) (length v2))
      (map 'vector #'* v1 v2))
     (t (error "No vector length adjustment yet"))
     ))
   (t (error "Bad operation data"))
   )
  )
(defun R/ (v1 v2)
  (cond
   ((and (fixnump v1) (fixnump v2))
      (/ (coerce (the fixnum v1) 'single-float)
	 (coerce (the fixnum v2) 'single-float)))
   ((and (fixnump v1) (floatp v2))
       (the single-float
	 (/ (coerce (the fixnum v1) 'single-float)
	    (the single-float v2))))
   ((and (floatp v1) (fixnump v2))
      (the single-float
	(/ (coerce (the single-float v1) 'single-float)
	   (the fixnum v2))))
   ((and (floatp v1) (floatp v2))
      (the single-float (/ (the single-float v1) (the single-float v2))))
   ((and (numberp v1) (numberp v2)) (/ v1 v2))
   ((and (numberp v1) (vectorp v2))
    (map 'vector #'/ (make-array (length v2) :initial-element v1) v2))
   ((and (numberp v2) (vectorp v1))
    (map 'vector #'/ (make-array (length v1) :initial-element v2) v1))
   ((and (vectorp v2) (vectorp v1))
    (cond
     ((eql (length v1) (length v2))
      (map 'vector #'/ v1 v2))
     (t (error "No vector length adjustment yet"))
     ))
   (t (error "Bad operation data"))
   )
  )



;(defmacro R+ (v1 v2)
;  (list 'R-op #'+ v1 v2)
;  )
;   
;(defmacro R- (v1 v2)
;  (list 'R-op #'- v1 v2)
;  )
;   
;(defmacro R* (v1 v2)
;  (list 'R-op #'* v1 v2)
;  )
;   
;(defmacro R/ (v1 v2)
;  (list 'R-op #'/ v1 v2)
;  )

;(defmacro R** (v1 v2)
;  (list 'R-op #'power v1 v2)
;  )

(defmacro R-elt (v i)
  (list 'aref v (list '1- i)))

;(defmacro R+ (v1 v2)
;  (list 'the 'single-float (list '+ v1 v2)))
;(defmacro R- (v1 v2)
;  (list 'the 'single-float (list '- v1 v2)))
;(defmacro R* (v1 v2)
;  (list 'the 'single-float (list '* v1 v2)))
;(defmacro R/ (v1 v2)
;  (list 'the 'single-float (list '/ v1 v2)))
   
(defun print (p)
  (princ p))

(defun sum (v)
  (if (vectorp v)
      (reduce #'+ v)
    v))

;;; oldfor is a strightforward implementation not used anymore

(defmacro oldfor (var vector body)
  (list 'map ''vector
	(list 'function 
	      (list 'lambda (list var) body))
	vector))

;;; this implementation detects the common idiom: for (j in a:b)
;;; where a and b are integer and converts it into a dotimes form

(defmacro for (var vector body)
  (if (and (listp vector) (eq 'seq (car vector)))
      (list 'dotimes 
	    (list 'z-i (list '1+ (list '- (third vector) (second vector))))
	    (list 'let (list (list var '(1+ z-i)))
		  (list 'declare (list 'fixnum var))
		  body))
    (list 'map ''vector
	  (list 'function 
		(list 'lambda (list var) body))
	  vector)
    ))

